;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "visual-fill-column" "2.7.1"
  "Fill-column for visual-line-mode."
  '((emacs "25.1"))
  :url "https://codeberg.org/joostkremers/visual-fill-column"
  :commit "9c0ecc2af21d3024a2a838c30d574e86265a52be"
  :revdesc "9c0ecc2af21d"
  :authors '(("Joost Kremers" . "joostkremers@fastmail.fm"))
  :maintainers '(("Joost Kremers" . "joostkremers@fastmail.fm")))
